local announcement_dialog = ...

local vars = DroidButtons.vars
local ui = DroidButtons.ui
local db = DroidButtons

local bool_map  = {[true] = 'ON', [false] = 'OFF'}
local state_map = {ON = true,  OFF = false}
local val_map   = {[true] = 1, [false] = 0}

local list2, list3 = dofile("templates.lua")

local region_type_radio_list = {'button', 'field'}
for i, v in ipairs(region_type_radio_list) do region_type_radio_list[v] = i end

local button_type_radio_list = {'standard', 'configurable'}
for i, v in ipairs(button_type_radio_list) do button_type_radio_list[v] = i end

local wInt = gkini.WriteInt
local wStr = gkini.WriteString

local function px(i) return tonumber(HUDSize(i/100, 0):match("(%d+)x")) end
local function py(i) return tonumber(HUDSize(0, i/100):match("x(%d+)")) end

local function check_number(i, min, max)
	i = tonumber(i); min = tonumber(min); max = tonumber(max)
	if     (not i)           then return false 
	elseif (min and i < min) then return false
	elseif (max and i > max) then return false
	end
	return true
end

local function ispluscommand(str) 
	assert(type(str) == "string", "bad argument #1 to 'ispluscommand' (string expected, got "..type(str)..")")
	if (str:sub(1,1) == "+") then return true end
	return false
end

local function make_dirty(ctl, num, min, max)
	if (num) then 
		local result = check_number(num, min or 0, max)
		if (result) then ctl.fgcolor = "255 255 255"; vars.TabState.BadValue = false
		else             ctl.fgcolor = "255 0 0"; vars.TabState.BadValue = true
		end
	end
	ui.Tabs:MakeDirty()
end


local function create_subdlg(ctrl)
	local dlg = iup.dialog{ctrl, border="NO", menubox="NO", resize="NO", bgcolor="0 0 0 0 *", expand = "YES", } --size = "FULL"}
	dlg.visible = "YES"
	return dlg
end

local function create_scrollbox(content)
	local scrollbox = iup.stationsublist{{}, control = "YES", expand = "YES",}
	iup.Append(scrollbox, create_subdlg(content))
	scrollbox[1] = 1
	return scrollbox
end

-- -------------------------------------------------------------------
-- Touch-to-Place Interface
-- -------------------------------------------------------------------
local show_placement_interface
local function show_drag()
	-- Can't Hide/Show DB dialog and work from the HUD.
	-- Attempting to Hide the droidbuttons maindialog, display the HUD, and chnage the input mode resulted in a CTD
	-- Workaround is to detach the display cbox (from the HUD), show it in a dialog, and then put it back on the HUD (when done).
	local closebtn, switchbtn
	local snap
	local grid
	local container
	local nudge_left, nudge_up, nudge_right, nudge_down
	local nudge_left_z, nudge_up_z, nudge_right_z, nudge_down_z
	local nudge_left_box, nudge_up_box, nudge_right_box, nudge_down_box

	local move_label_UD_size = "%5x%6"
	local move_label_LR_size = "%4x%8"

	closebtn    = iup.stationbutton{title = "Close"}
	switchbtn   = iup.stationbutton{title = "Switch", visible = "NO"}
	snap        = iup.stationtoggle{title = "Snap-to-Grid", value = db.snap_to_grid and "ON" or "OFF"}
	grid        = iup.text{value = db.grid_size, size = 50}
	nudge_left  = iup.stationbutton{title = "L", bgcolor = "0 0 0 0 *"}
	nudge_up    = iup.stationbutton{title = "U", bgcolor = "0 0 0 0 *"}
	nudge_right = iup.stationbutton{title = "R", bgcolor = "0 0 0 0 *"}
	nudge_down  = iup.stationbutton{title = "D", bgcolor = "0 0 0 0 *"}

	local nudge_z = {
		iup.zbox{
			nudge_up,
			iup.label{title = "Up", image = IMAGE_DIR.."arrow.png", size = move_label_UD_size},
			alignment = "ACENTER",
			all = "YES",
		},
		iup.zbox{
			nudge_down,
			iup.label{title = "Down", image = IMAGE_DIR.."arrow_down.png", size = move_label_UD_size},
			alignment = "ACENTER",
			all = "YES",
		},
		iup.zbox{
			nudge_left,
			iup.label{title = "Left", image = IMAGE_DIR.."arrow_left.png", size = move_label_LR_size},
			alignment = "ACENTER",
			all = "YES",
		},
		iup.zbox{
			nudge_right,
			iup.label{title = "Right", image = IMAGE_DIR.."arrow_right.png", size = move_label_LR_size},
			alignment = "ACENTER",
			all = "YES",
		},
	}

	local nudge_box = {
		iup.vbox{nudge_z[1]}, -- Up
		iup.vbox{nudge_z[2]}, -- Down
		iup.vbox{nudge_z[3]}, -- Left
		iup.vbox{nudge_z[4]}, -- Right
	}

	container = mdi2.pseudo_dialog{
		iup.vbox{
			nudge_box[1], -- Up
			iup.hbox{
				nudge_box[3], -- Left
				iup.pdarootframe{
					iup.vbox{
						snap,
						iup.hbox{iup.label{title = "Grid Size "}, grid},
						switchbtn,
						closebtn,
						alignment = "ACENTER",
						gap = 10,
					}
				},
				nudge_box[4], -- Right
				alignment = "ACENTER",
				gap = HUDSize(0.01),
			},
			nudge_box[2], -- Down
			alignment = "ACENTER",
			gap = HUDSize(0.01),
		},

		parent = db.cbox,
		cx = "ACENTER",
		cy = "ACENTER",
		resize = "NO",
		move = "NO",
		margin = "10x10",

	}

	iup.Detach(db.cbox)
	db.cbox.active = "YES"

	db.dragdlg = iup.dialog{
		iup.vbox{db.cbox,},
		topmost = "YES",
		fullscreen = "YES",
		bgcolor = "255 255 255 0 *",
		defaultesc = closebtn,
	}

	local switch = false
	function db.placement_arrows_toggle(show)
		switchbtn.visible = show and "YES" or "NO"
		if (not show and switch) then
			switchbtn:action()
		end
	end

	function db.placement_zorder()
		container:move_to_top()
	end

	function db.dragdlg:show_cb()
		snap:action()
		grid:action(nil, db.grid_size)
	end

	function switchbtn:action()
		local adj = 0
		for z = 1, 4 do
			local b
			if (switch) then -- if the button positions have been switched, then return them to normal
				b = z
			else
				b = z % 2
				if (b == 1) then adj = z end
				b = b + adj
			end

			iup.Detach(nudge_z[z])
			iup.Append(nudge_box[b], nudge_z[z])
		end
		iup.Refresh(db.dragdlg)
		switch = not switch
	end

	function closebtn:action()
		local timer = Timer()
		timer:SetTimeout(1, function()
			if (mdi2.active_pdlg) then
				mdi2.active_pdlg:highlight(nil, true)
				mdi2.active_pdlg = nil
			end
			HideDialog(db.dragdlg)
			iup.Detach(db.cbox)
			iup.Destroy(container)
			--iup.Refresh(db.dragdlg)
			iup.Destroy(db.dragdlg)
			db.cbox.active = "NO"
			iup.Append(HUD.pluginlayer, db.cbox)
			iup.Map(HUD.pluginlayer)
			db:CreateRegions(nil, nil, true) -- Create only the regions. Displays still present.
			ShowDialog(ui.MainDialog)
			db:save_buttondefs()
		end)
	end

	function snap:action()
		db.snap_to_grid = self.value == "ON" and true or false
		mdi2.snap_to_grid = db.snap_to_grid
		wInt(db.major, 'SnapToGrid', val_map[db.snap_to_grid])
	end

	function grid:action(c, after)
		db.grid_size = tonumber(after) or 0
		mdi2.grid_size = db.grid_size
		wInt(db.major, 'GridSize', db.grid_size)
	end

	function nudge_up:action()
		if (mdi2.active_pdlg) then
			mdi2.active_pdlg:highlight(nil, nil, true)
			mdi2.active_pdlg:move_and_resize(nil, ((db.snap_to_grid and db.grid_size or 1) * -1 * (switch and -1 or 1)), nil, nil, true, switch)
		end
	end
	function nudge_down:action()
		if (mdi2.active_pdlg) then
			mdi2.active_pdlg:highlight(nil, nil, true)
			mdi2.active_pdlg:move_and_resize(nil, ((db.snap_to_grid and db.grid_size or 1) * (switch and -1 or 1)), nil, nil, true, switch)
		end
	end
	function nudge_left:action()
		if (mdi2.active_pdlg) then
			mdi2.active_pdlg:highlight(nil, nil, true)
			mdi2.active_pdlg:move_and_resize(((db.snap_to_grid and db.grid_size or 1) * -1 * (switch and -1 or 1)), nil, nil, nil, true, switch)
		end
	end
	function nudge_right:action()
		if (mdi2.active_pdlg) then
			mdi2.active_pdlg:highlight(nil, nil, true)
			mdi2.active_pdlg:move_and_resize(((db.snap_to_grid and db.grid_size or 1) * (switch and -1 or 1)), nil, nil, nil, true, switch)
		end
	end



	iup.Map(db.dragdlg)
	HideDialog(ui.MainDialog)
	db:DestroyRegions(nil, true) -- Destroy only the regions.  Need displays for positioning
	ShowDialog(db.dragdlg)
end

-- -------------------------------------------------------------------
-- Edit Button Controls
-- -------------------------------------------------------------------
function ui:BuildButtonControls()
	local bc = {}
	local commands_zbox, commands_zbox_items
	local button_type_radio_items
	local standard_commands, configurable_commands
	local add_button
	
	-- Button region controls.  They will need to be populated later.
	bc.cmd1, bc.cmd1text, bc.cmd1list = list2{list = vars.commandlist, visible_items = 10, action = make_dirty}
	bc.cmd2, bc.cmd2text, bc.cmd2list = list2{list = vars.commandlist, visible_items = 10}
	bc.cmd2c, bc.cmd2ctext, bc.cmd2clist = list2{list = vars.commandlist, visible_items = 10, action = make_dirty, bgcolor = '55 122 150 255 *', bginactive = "128 128 128 255 *", active = "NO"}
	
	
	button_type_radio_items = {iup.stationradio{title = 'Standard', value = "ON"}, iup.stationradio{title = 'Configurable'}, gap = "%2"}
	bc.button_type_radio = iup.radio{iup.hbox(button_type_radio_items)}

	standard_commands = iup.vbox{
		iup.vbox{
			iup.label{title = "Button Command(s)"},
			iup.hbox{iup.label{title = 'Short-press: '}, iup.fill{}, bc.cmd1},
		},
		iup.hbox{iup.label{title = 'Long-press: '}, iup.fill{}, bc.cmd2},
		iup.hbox{iup.label{title = 'Long-press cancel: '}, iup.fill{}, bc.cmd2c},
		gap = "%3",
		expand = "NO",
	}
	
	bc.configurable_list = list3{expand = "YES", scrollbar = "VERTICAL"}--size = "500x500"}
	add_button = iup.stationbutton{title = "Add"}
	
	configurable_commands = iup.vbox{
		bc.configurable_list,
		iup.hbox{
			iup.fill{},
			add_button,
			iup.fill{},
		},
	}
	
	-- Layout ---------------------------------------------------------
	commands_zbox_items = {alignment = "NW"}
	-- Button region controls
	commands_zbox_items[1] = standard_commands
	-- Field region controls
	commands_zbox_items[2] = configurable_commands
	commands_zbox = iup.zbox(commands_zbox_items)


	bc.layout = iup.vbox{
		bc.button_type_radio,
		commands_zbox,
		gap = "%3",
	}

	-- Callbacks & Functions -----------------------------------------

	function add_button:action()
		local list = bc.configurable_list
		list:AddItem{title = "", command = ""}
		make_dirty()
		list:PopulateList()
		if (list:GetNumItems() == 5) then self.active = "NO" end
	end

	bc.cmd2:SetActionFunc(function(self, c, after) 
		local result = ispluscommand(after)
		bc.cmd2c:Activate(result)
		if (not result) then bc.cmd2c:Clear() end
		make_dirty()
	end)

	function bc.button_type_radio:set_value(v)
		self.value = button_type_radio_items[v]
		button_type_radio_items[v]:action()
	end

	local function radio_switch(index)
		commands_zbox.value = commands_zbox_items[index]
	end
	
	button_type_radio_items[1].action = function(self, v) 
		if (v) then make_dirty() end
		radio_switch(1)
	end
		
	button_type_radio_items[2].action = function(self, v)
		if (v) then make_dirty() end
		radio_switch(2)
	end
	
	bc.configurable_list:SetCreateItemFunc(function(item, index)
		local delete, move_up, move_down, title, cmd, layout
		move_size = "%3x%3"
		move_label_size = "%4x%4"
		delete = iup.stationbutton{title = "X", size = "%3x%5"}
		move_up = iup.stationbutton{title = "Up", size = move_size, fgcolor = "0 0 0 0 *", bgcolor = "0 0 0 0 *"}
		move_down = iup.stationbutton{title = "Down", size = move_size, fgcolor = "0 0 0 0 *", bgcolor = "0 0 0 0 *"}

		cmd = list2{name = "command", list = vars.commandlist_noplus, visible_items = 10, action = make_dirty}
		cmd.name = "command"
		cmd:SetText(item.command)
		title = iup.text{name = "title", value = item.title, size = cmd:GetWidth()}

		layout = iup.hbox{
			iup.vbox{
				iup.hbox{iup.label{title = "Title:"}, iup.fill{}, title, gap = 5},
				iup.hbox{iup.label{title = "Command:"}, iup.fill{}, cmd, gap = 5},
				gap = "%1",
				expand = "NO"
			},
			iup.vbox{
				iup.zbox{
					move_up,
					iup.label{title = "Up", image = IMAGE_DIR.."arrow.png", size = move_label_size},
					alignment = "ACENTER",
					all = "YES",
				},
				iup.fill{},
				iup.zbox{
					move_down,
					iup.label{title = "Down", image = IMAGE_DIR.."arrow_down.png", size = move_label_size},
					alignment = "ACENTER",
					all = "YES",
				},
			},
			iup.vbox{
				iup.fill{},
				delete,
				iup.fill{},
			},
			gap = 10,
			margin = "10x10",
		}

		function delete:action()
			bc.configurable_list:RemoveItem(index)
			add_button.active = "YES"
			make_dirty()
		end

		function move_up:action()
			bc.configurable_list:MoveItemUp(index)
			make_dirty()
		end
	                                      
		function move_down:action()
			bc.configurable_list:MoveItemDown(index)
			make_dirty()
		end
	                                      
		cmd:SetActionFunc(function(self, c, after)
			if (ispluscommand(after)) then return iup.IGNORE end
			iup.GetDialog(self):UpdateItem()
			make_dirty()
		end)
							

		function title:action(c, after)
			iup.GetDialog(self):UpdateItem()
			make_dirty()
		end

		return layout
	end)
	bc.configurable_list:SetItems{{title = "", command = ""}}
	bc.configurable_list:SetColors{'30 55 78 96', '42 74 96 96'} --, [0] = '65 100 127 255'}
	bc.configurable_list:PopulateList()
	
		
	return bc
end	


-- -------------------------------------------------------------------
-- Edit Field Controls
-- -------------------------------------------------------------------
function ui:BuildFieldControls()
	local fc = {fbc = {}, sensitivity= {}}
	-- Field region controls
	fc.xaxiscmd = iup.stationsublist(vars.joycmdlist)
	fc.yaxiscmd = iup.stationsublist(vars.joycmdlist)
	fc.cmd3, fc.cmd3text, fc.cmd3list = list2{list = vars.commandlist, visible_items = 10, action = make_dirty}
	fc.fbc.invertX = iup.stationtoggle{title = 'Invert X-axis'}
	fc.fbc.invertY = iup.stationtoggle{title = 'Invert Y-axis'}
	fc.fbc.mouselookX = iup.stationtoggle{title = 'Delta X'}
	fc.fbc.mouselookY = iup.stationtoggle{title = 'Delta Y'}
	fc.fbc.relative = iup.stationtoggle{title = 'Relative'}
	fc.fbc.lookspring = iup.stationtoggle{title = "Turret Lookspring"}
	fc.fbc.trackingcircles = iup.stationtoggle{title = "Touch Tracking Circles"}
	fc.sensitivity.curve = iup.stationsublist(vars.sensitivity_curve_list)
	fc.sensitivity.filter = iup.stationtoggle{title = 'Filtering'}
	fc.sensitivity.display = iup.label{title = "0.00x", alignment = 'ARIGHT'}
	fc.sensitivity.value = iup.canvas{size = "200x"..Font.Default, scrollbar = "HORIZONTAL", xmin = 1, xmax = 300, posx = 100, dx = Font.Default/2, border = "NO"}

	fc.layout = iup.vbox{
		iup.hbox{
			iup.vbox{
				iup.hbox{iup.label{title = "X-axis"}, iup.fill{}},
				iup.pdasubframebg{fc.xaxiscmd},
				fc.fbc.invertX,
				fc.fbc.mouselookX,
				gap = "%1"
			},
			iup.vbox{
				iup.hbox{iup.label{title = "Y-axis"}, iup.fill{}},
				iup.pdasubframebg{fc.yaxiscmd},
				fc.fbc.invertY,
				fc.fbc.mouselookY,
				gap = "%1"
			},
			gap = "%10",
		},
		iup.hbox{iup.fill{}, fc.fbc.relative, fc.fbc.lookspring, iup.fill{}, gap = "%5" },
		iup.hbox{iup.label{title = "Touch Command: "}, fc.cmd3},

		iup.vbox{
			iup.label{title = "Sensitivity"},
			iup.hbox{iup.label{title = "Curve"}, iup.pdasubframebg{fc.sensitivity.curve}, fc.sensitivity.filter, gap = "%3"},
			iup.hbox{fc.sensitivity.display, fc.sensitivity.value, gap = "%1"},
			gap = "%1",
		},
		iup.hbox{fc.fbc.trackingcircles, iup.fill{}},
		gap = "%3",
		expand = "NO",
	}
	
	function fc.xaxiscmd:action(t, i, v)
		if (i == tonumber(self.cursel)) then return iup.IGNORE end
		if (v == 1) then self.cursel = i; make_dirty() end
	end
	function fc.yaxiscmd:action(t, i, v)
		if (i == tonumber(self.cursel)) then return iup.IGNORE end
		if (v == 1) then self.cursel = i; make_dirty() end
	end

	for _, c in pairs(fc.fbc) do 
		function c:action(v) make_dirty() end 
	end

	function fc.sensitivity.curve:action(t, i, v)
		if (i == tonumber(self.cursel)) then return iup.IGNORE end
		if (v == 1) then self.cursel = i; make_dirty() end
	end
	function fc.sensitivity.filter:action(v) make_dirty() end

	function fc.sensitivity.value:scroll_cb(op, posx, posy) 
		-- called a PopulateFields() with op == nil to set display
		fc.sensitivity.display.title = string.format("%4.2fx", posx / 100)
		if (op) then make_dirty() end
	end

	return fc
end

-- -------------------------------------------------------------------
-- Edit Common Controls
-- -------------------------------------------------------------------
function ui:BuildCommonControls()
	local cc = {}
	-- Common region controls
	cc.xpos = iup.text{size = px(7.5)}
	cc.ypos = iup.text{size = px(7.5)}
	cc.zpos = iup.text{size = px(7.5), minnum = -99, maxnum = 99}
	cc.width = iup.text{size = px(7.5)}
	cc.height = iup.text{size = px(7.5)}
	cc.font = iup.text{size = px(7.5)}
	cc.fgcolor = iup.text{size = px(15)}
	cc.bgcolor = iup.text{size = px(15)}
	cc.isvisible = iup.stationtoggle{title = '', size = px(7.5)}
	cc.rounded = iup.stationtoggle{title = '', size = px(7.5)}

	cc.layout = iup.vbox{
		iup.hbox{
			iup.vbox{
				iup.hbox{iup.label{title = 'X Position: '}, iup.fill{}, cc.xpos,},
				iup.hbox{iup.label{title = 'Width: '}, iup.fill{}, cc.width,},
				iup.hbox{iup.label{title = 'Font Size: '}, iup.fill{}, cc.font,},
				gap = "%3"
			},
			iup.vbox{
				iup.hbox{iup.label{title = 'Y Position: '}, iup.fill{}, cc.ypos},
				iup.hbox{iup.label{title = 'Height: '}, iup.fill{}, cc.height},
				iup.hbox{iup.label{title = 'Visible '}, iup.fill{}, cc.isvisible},
				gap = "%3"
			},
			gap = "%3",
		},
		iup.hbox{iup.label{title = 'Z Position: '}, cc.zpos},
		iup.label{title = 'Z-Position allows you to place buttons above or below other buttons.  Smaller numbers are on top of larger numbers.  Negative values allowed.', wordwrap = "YES", size = Font.Default * 15},
		iup.hbox{iup.label{title = "Rounded Corners  ",}, cc.rounded},
		iup.hbox{iup.label{title = 'Foreground Color:  '}, cc.fgcolor},
		iup.hbox{iup.label{title = 'Background Color:  '}, cc.bgcolor},
		iup.vbox{
			iup.label{title = "Color codes should be in the form of \"R G B A\""},
			iup.label{title = "R: 0-255, G: 0-255, B: 0-255, A: 0-255"},
			iup.label{title = "Use blank to reset"},
		},
		iup.fill{},
		expand = "VERTICAL",
		gap = "%3"
	}

	do	
		local function func(self, c, after)
			if (c == 27 and after == "") then return end
			make_dirty(self, after, self.minnum, self.maxnum)
		end
		cc.xpos.action = func
		cc.ypos.action = func
		cc.zpos.action = func
		cc.width.action = func
		cc.height.action = func
		cc.font.action = func
	end
	cc.fgcolor.action = function(self, c, after) make_dirty() end
	cc.bgcolor.action = function(self, c, after) make_dirty() end
	function cc.isvisible:action(v) make_dirty() end
	function cc.rounded:action(v)
		local dcsr = db.defaults["bgcolor_button_rounded"]
		local dcs = db.defaults["bgcolor_button"]
		local cs = cc.bgcolor.value
		if (v == 1 and cs == dcs) then
			cc.bgcolor.value = dcsr
		elseif (v == 0 and cs == dcsr) then
			cc.bgcolor.value = dcs
		end
		make_dirty()
	end

	return cc
end

-- -------------------------------------------------------------------
-- Edit Region Interface
-- -------------------------------------------------------------------
function ui:BuildEditRegionInterface()
	local buttonindex -- set by PopulateFields() below
	local newbutton -- set by PopulateFields() below

	local editcontrols, editcontrols_scrollbox, editcontrols_content
	local editcontrols_zbox, editcontrols_zbox_items

	local bc, fc, cc -- table to hold button/field/common controls
	
	local title, type_radio, type_radio_items
	local positionbtn

	bc = ui:BuildButtonControls()
	fc = ui:BuildFieldControls()
	cc = ui:BuildCommonControls()

	title = iup.text{size = bc.cmd1text.size}
	
	type_radio_items = {iup.stationradio{title = 'Button', value = "ON"}, iup.stationradio{title = 'Field'}, gap = "%2"}
	type_radio = iup.radio{iup.hbox(type_radio_items)}

	positionbtn = iup.stationbutton{title = 'Position Regions'}

	-- Layout ---------------------------------------------------------
	editcontrols_zbox_items = {alignment = "NW"}
	-- Button region controls
	editcontrols_zbox_items[1] = bc.layout
	-- Field region controls
	editcontrols_zbox_items[2] = fc.layout
	editcontrols_zbox = iup.zbox(editcontrols_zbox_items)

	editcontrols_content = iup.vbox{
		iup.pdasubframebg{iup.hbox{iup.label{title = "Region Type: "}, type_radio, iup.fill{expand = "YES"}}},
		iup.vbox{
			iup.vbox{
				iup.hbox{iup.label{title = 'Title: '}, title},
				editcontrols_zbox,
				iup.fill{},
				gap = "%3"
			},
			cc.layout,
			gap = "%2",
			margin = "%1x%1"
		},
		gap = "%1",
	}
	editcontrols_content[1].border = string.format("%d %d %d %d", px(1), py(1), px(1), py(1)) -- left, top, right, bottom
	editcontrols_content[2].border = string.format("%d %d %d %d", px(1), py(1), px(1), py(1)) -- left, top, right, bottom

	editcontrols_scrollbox = create_scrollbox(editcontrols_content)
	editcontrols_scrollbox.active = "NO"
	editcontrols_scrollbox.bgcolor = "0 0 0 0 *"

	editcontrols = iup.pdasubsubsubframe2{
		iup.pdasubsubframebg{
			editcontrols_scrollbox,
			border = string.format("%d %d %d %d", px(0.5), py(1), px(0.5), py(1)) -- left, top, right, bottom
		},
	}
	-- ----------------------------------------------------------------

	local function reset_fields()
		cc.xpos.fgcolor = "255 255 255"
		cc.ypos.fgcolor = "255 255 255"
		cc.zpos.fgcolor = "255 255 255"
		cc.width.fgcolor = "255 255 255"
		cc.height.fgcolor = "255 255 255"
		cc.font.fgcolor = "255 255 255"
		cc.fgcolor.fgcolor = "255 255 255"
		cc.bgcolor.fgcolor = "255 255 255"

		vars.TabState.BadValue = false
	end
	
	function title:action(c, after) make_dirty() end

	function positionbtn:action()
		show_drag()
	end

	function type_radio:set_value(v)
		self.value = type_radio_items[v]
		type_radio_items[v]:action()
	end

	local function radio_switch(index)
		editcontrols_zbox.value = editcontrols_zbox_items[index]
		if (newbutton) then
			local rtype = region_type_radio_list[index]
			cc.width.value = db.defaults[rtype.."_width"]
			cc.height.value = db.defaults[rtype.."_height"]
			cc.bgcolor.value = db.defaults["bgcolor_"..rtype..(rtype == "button" and cc.rounded.value == "ON" and "_rounded" or "")]
			cc.xpos.value = db.xres / 2 - cc.width.value / 2
			cc.ypos.value = db.yres * 0.20
			iup.GetParent(cc.rounded).visible = rtype == "button" and "YES" or "NO"
			if (buttonindex ~= 0) then
				local buttondef = db.buttondefs[buttonindex]
				buttondef.type = rtype
				buttondef.x = cc.xpos.value
				buttondef.y = cc.ypos.value
				buttondef.w = cc.width.value
				buttondef.h = cc.height.value
				db:DeleteButton(buttonindex)
				db:CreateButton(buttonindex)
			end
		end
	end
	
	type_radio_items[1].action = function(self, v) radio_switch(1) end
		
	type_radio_items[2].action = function(self, v) radio_switch(2) end

	function editcontrols:PopulateFields(index)
		buttonindex = index
		local default = db.defaults
		local buttondef = db.buttondefs[index]

		if (buttondef) then
			newbutton = false
		else
			buttondef = {
				type = "button", 
				x = -1,  -- 0,
				y = -1, -- 0,
				w = default["button_width"], 
				h = default["button_height"], 
				font = default["font"],
				fgcolor = default["fgcolor"],
			}
			reset_fields()
			if (index ~= 0) then
				db.buttondefs[index] = buttondef
				db:CreateButton(index)
			end
			newbutton = true
		end
 
		-- Common region controls
		title.value = buttondef.title or ""
		cc.xpos.value = buttondef.x or 0
		cc.ypos.value = buttondef.y or 0
		cc.zpos.value = buttondef.z or default['zpos']
		cc.width.value = buttondef.w or default[buttondef.type.."_width"]
		cc.height.value = buttondef.h or default[buttondef.type.."_height"]
		cc.font.value = buttondef.font or default["font"]
		cc.fgcolor.value = buttondef.fgcolor or default["fgcolor"]
		cc.fgcolor.active = buttondef.type == "button" and "YES" or "NO"
		cc.bgcolor.value = buttondef.bgcolor or default["bgcolor_"..buttondef.type..(buttondef.rounded and buttondef.type == "button" and "_rounded" or "")]
		cc.isvisible.value = bool_map[buttondef.visible or true]
		cc.rounded.value = bool_map[buttondef.rounded or false]
		cc.rounded.visible = buttondef.type == "button" and "YES" or "NO"

		-- If it's a newbutton or a reset (index == 0) then all fields should be cleared
		-- If it's an existing button, then it shouldn't matter if the other region types controls are populated
		-- Button region controls
		bc.cmd1text.value = buttondef.command1 or ''
		bc.cmd2text.value = buttondef.command2 or ''
		bc.cmd2ctext.value = buttondef.command2cancel or ''
		bc.cmd2c:Activate(buttondef.command2cancel)
		bc.button_type_radio:set_value(button_type_radio_list[buttondef.button_type or "standard"]) --TODO update data structure or leave default?
		bc.configurable_list:SetItems(buttondef.commands or {})
		bc.configurable_list:PopulateList()

		-- Field region controls
		fc.xaxiscmd.value = vars.joycmdlist2[buttondef.command1] or vars.joycmdlist2['NONE']
		fc.xaxiscmd.cursel = fc.xaxiscmd.value
		fc.yaxiscmd.value = vars.joycmdlist2[buttondef.command2] or vars.joycmdlist2['NONE']
		fc.yaxiscmd.cursel = fc.yaxiscmd.value
		fc.cmd3text.value = buttondef.command3 or ''
		fc.fbc.invertX.value = bool_map[buttondef.invertX or false]
		fc.fbc.invertY.value = bool_map[buttondef.invertY or false]
		fc.fbc.mouselookX.value = bool_map[buttondef.mouselookX or false]
		fc.fbc.mouselookY.value = bool_map[buttondef.mouselookY or false]
		fc.fbc.relative.value = bool_map[buttondef.relative or false]
		fc.fbc.lookspring.value = bool_map[not(buttondef.lookspring) or false]
		fc.fbc.trackingcircles.value = bool_map[buttondef.trackingcircles or true]

		local cv = buttondef.Scurve or 256 -- default to unfiltered linear curve
		fc.sensitivity.filter.value = bool_map[(cv < 256) or false]
		cv = cv + 1 - (cv < 256 and 0 or 256)
		fc.sensitivity.curve.value = cv
		fc.sensitivity.curve.cursel = cv
		fc.sensitivity.value.posx = buttondef.Svalue or 100
		fc.sensitivity.value:scroll_cb(nil, fc.sensitivity.value.posx)

		type_radio:set_value(region_type_radio_list[buttondef.type])
		local type_idx = region_type_radio_list[buttondef.type]
		for i, ctl in ipairs(type_radio_items) do
			if (type_idx == i or newbutton) then ctl.active = "YES" else ctl.active = "NO" end
		end

		-- index == 0 designates a deleted record.  Disable all controls.
		editcontrols_scrollbox.active = index == 0 and "NO" or "YES"
		editcontrols_scrollbox.bgcolor = index == 0 and "0 0 0 0 *" or "255 255 255 255 *"
		editcontrols_scrollbox.scroll = "TOP"
	end
	
	function editcontrols:Save()
		if (vars.TabState.BadValue) then return false end
		local recreate_region
		local buttondef = db.buttondefs[buttonindex]
		local newtitle = buttondef.title ~= title.value
		local bgcolor_button_key = "bgcolor_"..type_radio.value.title:lower()..(cc.rounded.value == "ON" and type_radio.value.title:lower() == "button" and "_rounded" or "")

		local function store_value(k, v, keep_region)
			if (buttondef[k] ~= v and not keep_region) then recreate_region = true end
			-- don't save these so globals can affect them.
			if (k == "bgcolor" and v == nil) then
				buttondef[k] = db.defaults[bgcolor_button_key]
				return
			elseif(k == "bgcolor" and v == db.defaults[bgcolor_button_key]) then return
			end

			if (k == "fgcolor" and v == nil) then
				buttondef[k] = db.defaults["fgcolor"]
				return
			elseif (k == "fgcolor" and v == db.defaults["fgcolor"]) then return
			end

			buttondef[k] = v
		end

		store_value('type', type_radio.value.title:lower(), true)
		store_value('title', title.value or '', true)
		store_value('x', tonumber(cc.xpos.value), true)
		store_value('y', tonumber(cc.ypos.value), true)
		store_value('z', tonumber(cc.zpos.value), false)
		store_value('w', tonumber(cc.width.value), true)
		store_value('h', tonumber(cc.height.value), true)
		store_value('font', tonumber(cc.font.value), true)
		store_value('fgcolor', cc.fgcolor.value:match("^((%d+ %d+ %d+)%s*(%d*))"), false)
		store_value('bgcolor', cc.bgcolor.value:match("^((%d+ %d+ %d+)%s*(%d*))"), false)
		store_value('visible', state_map[cc.isvisible.value], true)
		store_value('rounded', state_map[cc.rounded.value],false)

		if (buttondef.type == 'button') then
			--local ipc = ispluscommand(buttondef.command1 or "")
			--if ((ipc and cmd1text.value ~= buttondef.command1) or (not ipc and ispluscommand(cmd1text.value))) then recreate_region = true end

			store_value('button_type', bc.button_type_radio.value.title:lower(), false)
			if (buttondef.button_type == 'standard') then
				store_value('command1', bc.cmd1text.value ~= '' and bc.cmd1text.value or nil, not (ispluscommand(buttondef.command1 or "") or ispluscommand(bc.cmd1text.value)))
				store_value('command2', bc.cmd2text.value ~= '' and bc.cmd2text.value or nil, true)
				store_value('command2cancel', bc.cmd2ctext.value ~= '' and bc.cmd2ctext.value or nil, true)
				store_value('commands', nil, true)
				
			elseif (buttondef.button_type == 'configurable') then
				store_value('command1', nil, not (ispluscommand(buttondef.command1 or "")))
				store_value('command2', nil, true)
				store_value('command2cancel', nil, true)
				local button_items = bc.configurable_list:GetItems()
				store_value('commands', button_items, true)
				buttondef.commands.current = #button_items > 0 and 1 or nil -- always reset the current command in case of changes/deletions
			end

		elseif (buttondef.type == 'field') then
			store_value('command1', vars.joycmdlist3[tonumber(fc.xaxiscmd.value)] ~= "NONE" and vars.joycmdlist3[tonumber(fc.xaxiscmd.value)] or nil)
			store_value('command2', vars.joycmdlist3[tonumber(fc.yaxiscmd.value)] ~= "NONE" and vars.joycmdlist3[tonumber(fc.yaxiscmd.value)] or nil)
			store_value('command3', fc.cmd3text.value ~= '' and fc.cmd3text.value or nil)

			for p, c in pairs(fc.fbc) do
				if (p == 'lookspring') then -- TODO If more exceptions are needed, consider making a table for each one with a true/false value to modify the result
					store_value(p, not(state_map[c.value]))
				else
					store_value(p, state_map[c.value])
				end
			end
			
			store_value('Scurve', tonumber(fc.sensitivity.curve.value) - 1 + (state_map[fc.sensitivity.filter.value] and 0 or 1)*256)
			store_value('Svalue', tonumber(fc.sensitivity.value.posx))
		end

		db:save_buttondefs()
		if (recreate_region) then
			db:DeleteButton(buttonindex) -- Not using drag-and-drop here.  delete the display and the region and then recreate
			db:CreateButton(buttonindex)
		else
			db:UpdateButton(buttonindex)
		end
		if (newbutton or newtitle) then
			return function() ui.ButtonsTab:PopulateTab(buttonindex) end
		end

		return true
	end

	function editcontrols:Undo()
		reset_fields()
		if (newbutton) then
			-- Treat undoing a newbutton as a delete
			db:DeleteButton(buttonindex)
			table.remove(db.buttondefs, buttonindex)
			return function() ui.ButtonsTab:PopulateTab(0) end
		else
			db:UpdateButton(buttonindex, nil, true, false)
			self:PopulateFields(buttonindex)
		end
	end
	-- ----------------------------------------------------------------

	return editcontrols, positionbtn

end
